@@ -12,7 +12,9 @@
 
 #define RUBY
 
-/* #include "config.h" */
+/* define EUC/SJIS for default kanji-code */
+#define EUC
+#undef SJIS
 
 /* define USE_DLN to load object file(.o). */
 #ifdef HAVE_A_OUT_H
